#include <Windows.h>

//#include "Resource.h"
#include "BonTuner.h"

#define CREATE_MUTEX_NAME L"Global\\BonDriver_HDUS_Create"

#define MAX_BUFF_COUNT 500

#pragma warning( disable : 4273 )
extern "C" __declspec(dllexport) IBonDriver * CreateBonDriver()
{
	return (CBonTuner::m_pThis)? CBonTuner::m_pThis : ((IBonDriver *) new CBonTuner);
}
#pragma warning( default : 4273 )

CBonTuner * CBonTuner::m_pThis = NULL;
HINSTANCE CBonTuner::m_hModule = NULL;


CBonTuner::CBonTuner()
{
	m_pThis = this;

	m_dwCurSpace = 0;
	m_dwCurChannel = 0;

	m_hOnStreamEvent = NULL;
	m_LastBuff = NULL;
	m_dwRecv1Sec = 0;
	m_dwBuffOffset = 0;

	m_pcDShow = NULL;

	::InitializeCriticalSection(&m_CriticalSection);

	WCHAR strExePath[512] = L"";
	GetModuleFileName(m_hModule, strExePath, 512);
	m_strExePath = strExePath;

	wstring strIniPath;
	strIniPath = strExePath;
	strIniPath+=L".ini";

	m_bNoUHF = GetPrivateProfileInt( L"Set", L"NoUHF", 0, strIniPath.c_str() );
	m_bNoCATV = GetPrivateProfileInt( L"Set", L"NoCATV", 0, strIniPath.c_str() );

	m_dLevel1 = 0;

	m_hMutex = NULL;

	m_hCriticalEvent = ::CreateEvent(NULL, FALSE, TRUE, CREATE_MUTEX_NAME);
	SetEvent(m_hCriticalEvent);
}

CBonTuner::~CBonTuner()
{
	CloseTuner();

	::DeleteCriticalSection(&m_CriticalSection);

	if( m_hCriticalEvent == NULL ){
		::CloseHandle(m_hCriticalEvent);
		m_hCriticalEvent = NULL;
	}

	m_pThis = NULL;
}

const BOOL CBonTuner::OpenTuner(void)
{
	if( WaitForSingleObject( m_hCriticalEvent, INFINITE ) != WAIT_OBJECT_0 ){
		if( m_hCriticalEvent != NULL ) SetEvent(m_hCriticalEvent);
		return FALSE;
	}

	CloseTuner();

	m_pcDShow = new CDShowMain;
	m_pcDShow->SetModulePath(m_strExePath);

	wstring strDevPath = L"";
	if( m_pcDShow->GetEmptyDevicePath(strDevPath) == FALSE ){
		if( m_hCriticalEvent != NULL ) SetEvent(m_hCriticalEvent);
		return FALSE;
	}
	//Cxg
	m_hOnStreamEvent = ::CreateEvent(NULL, FALSE, FALSE, NULL);

	m_dwRecv1Sec = 0;
	m_dwBuffOffset = 0;


	if( m_pcDShow->OpenTuner(strDevPath, RecvProc, this) == FALSE ){
		if( m_hCriticalEvent != NULL ) SetEvent(m_hCriticalEvent);
		return FALSE;
	}
	wstring strMutex = L"Global\\";
	int pos = (int)strDevPath.find(L"#");
	if( pos != string::npos ){
		pos = (int)strDevPath.find(L"#",pos+1);
	}
	int pos2 = (int)strDevPath.rfind(L"#");
	strMutex += strDevPath.substr(pos+1, pos2-(pos+1));

	m_hMutex = CreateMutex( NULL,TRUE,strMutex.c_str() );

	if( m_hCriticalEvent != NULL ) SetEvent(m_hCriticalEvent);
	//`l
	SetChannel(0,0);

	return TRUE;
}

void CBonTuner::CloseTuner(void)
{
	if( m_pcDShow != NULL ){
		m_pcDShow->CloseTuner();
		Sleep(100);
		SAFE_DELETE(m_pcDShow);
	}

	//obt@
	::EnterCriticalSection(&m_CriticalSection);
	if( m_LastBuff != NULL ){
		SAFE_DELETE(m_LastBuff);
	}
	for( int i=0; i<(int)m_TsBuff.size(); i++ ){
		SAFE_DELETE(m_TsBuff[i])
	}
	m_TsBuff.clear();
	m_dwBuffOffset = 0;
	::LeaveCriticalSection(&m_CriticalSection);

	if( m_hMutex != NULL ){
		::ReleaseMutex(m_hMutex);
		::CloseHandle(m_hMutex);
	}
	m_hMutex = NULL;
}

const BOOL CBonTuner::SetChannel(const BYTE bCh)
{
	return TRUE;
}

const float CBonTuner::GetSignalLevel(void)
{
	if( WaitForSingleObject( m_hCriticalEvent, INFINITE ) != WAIT_OBJECT_0 ){
		if( m_hCriticalEvent != NULL ) SetEvent(m_hCriticalEvent);
		return 0;
	}
	float fBitRate = 0;
	fBitRate = ((float)m_dwRecv1Sec/(1024*1024))*8;

	m_pcDShow->GetLevel(&m_dLevel1, NULL );

	if( m_hCriticalEvent != NULL ) SetEvent(m_hCriticalEvent);
	if( m_dLevel1 == -2 ){
		return fBitRate;
	}else{
		return (float)m_dLevel1;
	}
}

const DWORD CBonTuner::WaitTsStream(const DWORD dwTimeOut)
{
	if( m_hOnStreamEvent == NULL ){
		return WAIT_ABANDONED;
	}
	// CxgVOiԂɂȂ̂҂
	const DWORD dwRet = ::WaitForSingleObject(m_hOnStreamEvent, (dwTimeOut)? dwTimeOut : INFINITE);

	switch(dwRet){
		case WAIT_ABANDONED :
			// `[iꂽ
			return WAIT_ABANDONED;

		case WAIT_OBJECT_0 :
		case WAIT_TIMEOUT :
			// Xg[擾\
			return dwRet;

		case WAIT_FAILED :
		default:
			// O
			return WAIT_FAILED;
	}
}

const DWORD CBonTuner::GetReadyCount(void)
{
	DWORD dwCount = 0;
	::EnterCriticalSection(&m_CriticalSection);
	dwCount = (DWORD)m_TsBuff.size();
	::LeaveCriticalSection(&m_CriticalSection);
	return dwCount;
}

const BOOL CBonTuner::GetTsStream(BYTE *pDst, DWORD *pdwSize, DWORD *pdwRemain)
{
	BYTE *pSrc = NULL;

	if(GetTsStream(&pSrc, pdwSize, pdwRemain)){
		if(*pdwSize){
			::CopyMemory(pDst, pSrc, *pdwSize);
		}
		return TRUE;
	}
	
	return FALSE;
}

const BOOL CBonTuner::GetTsStream(BYTE **ppDst, DWORD *pdwSize, DWORD *pdwRemain)
{
	static BYTE bXORtable[] = {
		0x6D, 0x4E, 0xF0, 0x1E, 0x59, 0x8B, 0xB2, 0xF0, 0x31, 0xF9, 0xD0, 0x85, 0xDD, 0x84, 0x4A, 0x6C,
		0x15, 0x1C, 0x1C, 0x0A, 0x5B, 0x61, 0x6F, 0x1A, 0xAD, 0x60, 0x46, 0xF2, 0xD5, 0x03, 0x91, 0xC7,
		0xB7, 0xDE, 0x68, 0x9A, 0x07, 0x47, 0x62, 0x5E, 0x39, 0xD6, 0x7E, 0x45, 0x10, 0x9D, 0x1A, 0xC4,
		0x4C, 0x7F, 0xB7, 0x03, 0xCC, 0xAD, 0x72, 0x6C, 0xE9, 0x1B, 0x85, 0xA8, 0xEE, 0x4B, 0xA9, 0x3A,
		0xFF, 0xF1, 0x3B, 0x16, 0xE0, 0x59, 0x4C, 0x3A, 0xD5, 0x4A, 0x14, 0x95, 0x0A, 0xD4, 0x25, 0x23,
		0xF6, 0x2F, 0xE3, 0x57, 0x35, 0xCE, 0x52, 0x84, 0x11, 0xA3, 0xCB, 0x34, 0xA7, 0x15, 0x60, 0xE1,
		0x98, 0x71, 0xF4, 0x1F, 0xF7, 0xBD, 0x22, 0x4B, 0x61, 0x8A, 0xD6, 0xAF, 0x24, 0x27, 0xC3, 0xA7,
		0x07, 0x9B, 0xA5, 0x57, 0x83, 0x80, 0x58, 0x07, 0x1D, 0xDE, 0x0A, 0xD8, 0xB9, 0x0A, 0xBE, 0xAC,
		0xA9, 0x76, 0x00, 0x1C, 0x23, 0x9F, 0x82, 0x79, 0x41, 0x56, 0xC8, 0x83, 0xC3, 0x38, 0x37, 0x4B,
		0x5A, 0xED, 0xC9, 0xD5, 0xF2, 0x87, 0x02, 0xC3, 0x79, 0xA0, 0x61, 0xD9, 0x63, 0x50, 0xCA, 0x72,
		0xC2, 0x51, 0x40, 0x98, 0xD9, 0x8B, 0xED, 0xA9, 0xA1, 0x4C, 0xA5, 0xE7, 0xAF, 0xF2, 0x08, 0x4D,
		0x82, 0xF8, 0xD5, 0x02, 0x61, 0xD8, 0xC2, 0xE1
	};
	static int iXORMode = -1;

	::EnterCriticalSection(&m_CriticalSection);
	if( m_LastBuff != NULL ){
		SAFE_DELETE(m_LastBuff);
	}
	::LeaveCriticalSection(&m_CriticalSection);
	BOOL bRet = TRUE;
	if( m_TsBuff.size() != 0 ){
		::EnterCriticalSection(&m_CriticalSection);
		m_LastBuff = m_TsBuff[0];
		m_TsBuff.erase( m_TsBuff.begin() );
		::LeaveCriticalSection(&m_CriticalSection);

		if( iXORMode == -1 ){
			int iReadCount = m_LastBuff->dwSize / 188;
			for( int i=0; i<iReadCount; i++ ){
				unsigned short usPID = ((unsigned short)(m_LastBuff->pbBuff[(i*188)+1]&0x1F))<<8 | m_LastBuff->pbBuff[(i*188)+2];
				if( usPID == 0x0000 ){
					if( m_LastBuff->pbBuff[(i*188)+4] == 0x00){
						iXORMode = 0;
					}else if( m_LastBuff->pbBuff[(i*188)+4] == 0x6D){
						iXORMode = 1;
					}
					break;
				}
			}
		}
		if( iXORMode == 1 ){
			int iReadCount = m_LastBuff->dwSize / 188;
			for( int i=0; i<iReadCount; i++ ){
				for( int j=0; j<184; j++ ){
					m_LastBuff->pbBuff[(i*188)+4+j] ^= bXORtable[j];
				}
			}
		}
		if( iXORMode != -1 ){
			*pdwSize = m_LastBuff->dwSize;
			*ppDst = m_LastBuff->pbBuff;
			*pdwRemain = (DWORD)m_TsBuff.size();
		}else{
			*pdwSize = 0;
			*pdwRemain = 0;
			bRet = FALSE;
		}
	}else{
		*pdwSize = 0;
		*pdwRemain = 0;
		bRet = FALSE;
	}
	return bRet;
}

void CBonTuner::PurgeTsStream(void)
{
	//obt@
	::EnterCriticalSection(&m_CriticalSection);
	if( m_LastBuff != NULL ){
		SAFE_DELETE(m_LastBuff);
	}
	for( int i=0; i<(int)m_TsBuff.size(); i++ ){
		SAFE_DELETE(m_TsBuff[i])
	}
	m_TsBuff.clear();
	m_dwBuffOffset = 0;
	::LeaveCriticalSection(&m_CriticalSection);

	m_dwRecv1Sec = 0;
}

LPCTSTR CBonTuner::GetTunerName(void)
{
	return TEXT("HDUS");
}

const BOOL CBonTuner::IsTunerOpening(void)
{
	return FALSE;
}
	
LPCTSTR CBonTuner::EnumTuningSpace(const DWORD dwSpace)
{
	switch(dwSpace){
		case 0:
			return TEXT("nfWiUHFj");
		case 1:
			return TEXT("nfWiP[uj");
		default:
			return NULL;
	}
}

LPCTSTR CBonTuner::EnumChannelName(const DWORD dwSpace, const DWORD dwChannel)
{
	static const TCHAR aszChannelName[][16] = 
	{
		TEXT("13ch"), TEXT("14ch"), TEXT("15ch"), TEXT("16ch"), TEXT("17ch"), TEXT("18ch"), TEXT("19ch"),
		TEXT("20ch"), TEXT("21ch"), TEXT("22ch"), TEXT("23ch"), TEXT("24ch"), TEXT("25ch"), TEXT("26ch"), TEXT("27ch"), TEXT("28ch"), TEXT("29ch"),
		TEXT("30ch"), TEXT("31ch"), TEXT("32ch"), TEXT("33ch"), TEXT("34ch"), TEXT("35ch"), TEXT("36ch"), TEXT("37ch"), TEXT("38ch"), TEXT("39ch"), 
		TEXT("40ch"), TEXT("41ch"), TEXT("42ch"), TEXT("43ch"), TEXT("44ch"), TEXT("45ch"), TEXT("46ch"), TEXT("47ch"), TEXT("48ch"), TEXT("49ch"), 
		TEXT("50ch"), TEXT("51ch"), TEXT("52ch"), TEXT("53ch"), TEXT("54ch"), TEXT("55ch"), TEXT("56ch"), TEXT("57ch"), TEXT("58ch"), TEXT("59ch"), 
		TEXT("60ch"), TEXT("61ch"), TEXT("62ch")
	};
	static const TCHAR aszCableChannelName[][16] = 
	{
		TEXT("C13ch"), TEXT("C14ch"), TEXT("C15ch"), TEXT("C16ch"), TEXT("C17ch"), TEXT("C18ch"), TEXT("C19ch"),
		TEXT("C20ch"), TEXT("C21ch"), TEXT("C22ch"), TEXT("C23ch"), TEXT("C24ch"), TEXT("C25ch"), TEXT("C26ch"), TEXT("C27ch"), TEXT("C28ch"), TEXT("C29ch"),
		TEXT("C30ch"), TEXT("C31ch"), TEXT("C32ch"), TEXT("C33ch"), TEXT("C34ch"), TEXT("C35ch"), TEXT("C36ch"), TEXT("C37ch"), TEXT("C38ch"), TEXT("C39ch"), 
		TEXT("C40ch"), TEXT("C41ch"), TEXT("C42ch"), TEXT("C43ch"), TEXT("C44ch"), TEXT("C45ch"), TEXT("C46ch"), TEXT("C47ch"), TEXT("C48ch"), TEXT("C49ch"), 
		TEXT("C50ch"), TEXT("C51ch"), TEXT("C52ch"), TEXT("C53ch"), TEXT("C54ch"), TEXT("C55ch"), TEXT("C56ch"), TEXT("C57ch"), TEXT("C58ch"), TEXT("C59ch"), 
		TEXT("C60ch"), TEXT("C61ch"), TEXT("C62ch"), TEXT("C63ch")
	};
	if( dwSpace == 0 && m_bNoUHF == FALSE ){
		if( dwChannel <= 62-13 ){
			return aszChannelName[dwChannel];
		}
	}else if( dwSpace == 1 && m_bNoCATV == FALSE ){
		if( dwChannel <= 63-13 ){
			return aszCableChannelName[dwChannel];
		}
	}

	return NULL;
}

const BOOL CBonTuner::SetChannel(const DWORD dwSpace, const DWORD dwChannel)
{
	if(!EnumChannelName(dwSpace, dwChannel))return FALSE;

	//gϊ
	int iFreq = 0;
	if( dwSpace == 0 ){
		//UHF
		iFreq = 473000 + 6000 * dwChannel;
	}else{
		//P[u
		if( dwChannel <= 22-13 ){
			iFreq = 111000 + 6000 * dwChannel;
		}else{
			iFreq = 225000 + 6000 * (dwChannel-10);
		}
	}
	//CHZbg
	if( m_pcDShow != NULL ){
		//`l؊ɑƂ܂H
		if( WaitForSingleObject( m_hCriticalEvent, INFINITE ) != WAIT_OBJECT_0 ){
			if( m_hCriticalEvent != NULL ) SetEvent(m_hCriticalEvent);
			return FALSE;
		}

		m_pcDShow->SetChFreq( iFreq );

		if( m_hCriticalEvent != NULL ) SetEvent(m_hCriticalEvent);


		m_dwCurSpace = dwSpace;
		m_dwCurChannel = dwChannel;
	}

	PurgeTsStream();

	return TRUE;
}
	
const DWORD CBonTuner::GetCurSpace(void)
{
	return m_dwCurSpace;
}

const DWORD CBonTuner::GetCurChannel(void)
{
	return m_dwCurChannel;
}

void CBonTuner::Release()
{
	delete this;
}

int CALLBACK CBonTuner::RecvProc(void* pParam, BYTE* pbData, DWORD dwSize)
{
	CBonTuner* pSys = (CBonTuner*)pParam;
	static DWORD dwRecvSize = 0;
	static DWORD dwTickCount = 0;

	if( dwSize > 0 ){
		DWORD dwCopySize = 0;
		BOOL bBuffFull = FALSE;

		if( BUFF_SIZE > pSys->m_dwBuffOffset+dwSize){
			dwCopySize = dwSize;
		}else{
			dwCopySize = BUFF_SIZE-pSys->m_dwBuffOffset;
			bBuffFull = TRUE;
		}
		memcpy(pSys->m_bRecvBuff+pSys->m_dwBuffOffset, pbData, dwCopySize);

		if( bBuffFull == TRUE ){
			TS_DATA* pItem = new TS_DATA;
			pItem->dwSize = BUFF_SIZE;
			pItem->pbBuff = new BYTE[BUFF_SIZE];
			memcpy(pItem->pbBuff, pSys->m_bRecvBuff, pItem->dwSize);

			::EnterCriticalSection(&pSys->m_CriticalSection);
			if( pSys->m_TsBuff.size() > MAX_BUFF_COUNT ){
				while( pSys->m_TsBuff.size() > MAX_BUFF_COUNT ){
					SAFE_DELETE(pSys->m_TsBuff[0]);
					pSys->m_TsBuff.erase( pSys->m_TsBuff.begin() );
				}
			}
			pSys->m_TsBuff.push_back(pItem);
			::LeaveCriticalSection(&pSys->m_CriticalSection);
			SetEvent(pSys->m_hOnStreamEvent);

			pSys->m_dwBuffOffset = dwSize-dwCopySize;
			if( pSys->m_dwBuffOffset > 0 ){
				memcpy(pSys->m_bRecvBuff, pbData+dwCopySize, pSys->m_dwBuffOffset);
			}
		}else{
			pSys->m_dwBuffOffset+=dwSize;
		}

		//rbg[gvZp
		DWORD dwTick = GetTickCount();
		if( dwTick - dwTickCount > 1000 ){
			pSys->m_dwRecv1Sec = dwRecvSize + dwSize;
			dwTickCount = dwTick;
			dwRecvSize = 0;
		}else if( dwTick < dwTickCount ){
			dwTickCount = dwTick;
			dwRecvSize = 0;
		}else{
			dwRecvSize += dwSize;
		}
	}
	return 0;
}
