#include "DShowMain.h"

#include <initguid.h>
#include <objbase.h>
#include <math.h>

#define DVBT_NP_NAME L"Microsoft DVBT Network Provider"
#define HDUS_SOURCE_NAME L"SKNET HDTV BDA Digital Tuner_0"
#define HDUS_COMPONENT_NAME L"SKNET HDTV BDA Digital Capture_0"
#define DVBT_TSPACE_NAME L"DC-DVBt"
#define MPEG2_DEMULTI_NAME L"MPEG-2 Demultiplexer"

static const GUID CLSID_PropSet
 = {0x9E1781E1,0x9CB1,0x4407,{0xBB,0xCE,0x54,0x26,0xC8,0xD0,0x0A,0x4B}};

#define _LOGOUT

static void err_translator(unsigned code, EXCEPTION_POINTERS *)
{
	throw code;
}

CDShowMain::CDShowMain(void)
{
	CoInitialize(NULL);
	m_bCreate = FALSE;

//	m_pcDmp = NULL;
	m_pcTee = NULL;

	m_pGraphBuilder = NULL;
	m_pMediaControl = NULL;
	m_pNetworkProvider = NULL;
	m_pTunerDevice = NULL;
	m_pCaptureDevice = NULL;
	m_pInfPinTee = NULL;
//	m_pDump = NULL;
	m_pMpeg2Demulti = NULL;
	m_pTIF = NULL;
	m_pTuningSpace = NULL;

	m_iLastChFreq = -1;

}

CDShowMain::~CDShowMain(void)
{
	if( m_bCreate != FALSE ){
		CloseTuner();
	}
	CoUninitialize();
}

void CDShowMain::SetModulePath(wstring strModulePath)
{
	m_strLogPath = strModulePath;
	m_strLogPath += L".log";
}

BOOL CDShowMain::EnumBDADevice(vector<DEVICE_INFO_DATA>* pDevList)
{
	HRESULT hr = S_OK;
	CComPtr <IMoniker> pIMoniker;
	CComPtr <IEnumMoniker> pIEnumMoniker;
	CComPtr <ICreateDevEnum> pICreateDevEnum;

	REFCLSID clsid = KSCATEGORY_BDA_NETWORK_TUNER;

	hr = pICreateDevEnum.CoCreateInstance(CLSID_SystemDeviceEnum);
	if (FAILED (hr)){
#ifdef _LOGOUT
		OutputLog("CoCreateInstance CLSID_SystemDeviceEnum Err 0x%x", hr);
#endif
		return FALSE;
	}
	hr = pICreateDevEnum->CreateClassEnumerator(KSCATEGORY_BDA_NETWORK_TUNER, &pIEnumMoniker, 0);
	if (hr == S_OK){
		ULONG cFetched;
		while(pIEnumMoniker->Next(1, &pIMoniker, &cFetched) == S_OK){
			IPropertyBag *pPropBag;
			hr = pIMoniker->BindToStorage(0, 0, IID_IPropertyBag, 
				(void **)&pPropBag);
			if (SUCCEEDED(hr))
			{
				DEVICE_INFO_DATA Item;
				VARIANT varName;
				varName.vt = VT_BSTR;
				hr = pPropBag->Read(L"FriendlyName", &varName, 0);
				if (SUCCEEDED(hr)){
					Item.strFriendlyName = varName.bstrVal;
#ifdef _LOGOUT
{
	int iLen = 0;
	iLen = WideCharToMultiByte( 932, 0, Item.strFriendlyName.c_str(), -1, NULL, 0, NULL, NULL );
	char* pszBuff = new char[iLen+1];
	ZeroMemory(pszBuff, sizeof(char)*(iLen+1));
	WideCharToMultiByte( 932, 0, Item.strFriendlyName.c_str(), -1, pszBuff, iLen, NULL, NULL );

	OutputLog("FriendlyName:%s",pszBuff);
	delete[] pszBuff;
}
#endif
				}
				VariantClear(&varName);

				varName.vt = VT_BSTR;
				hr = pPropBag->Read(L"Description", &varName, 0);
				if (SUCCEEDED(hr)){
					Item.strDescription = varName.bstrVal;
				}
				VariantClear(&varName);

				varName.vt = VT_BSTR;
				hr = pPropBag->Read(L"DevicePath", &varName, 0);
				if (SUCCEEDED(hr)){
					Item.strDevicePath = varName.bstrVal;
#ifdef _LOGOUT
	int iLen = 0;
	iLen = WideCharToMultiByte( 932, 0, Item.strDevicePath.c_str(), -1, NULL, 0, NULL, NULL );
	char* pszBuff = new char[iLen+1];
	ZeroMemory(pszBuff, sizeof(char)*(iLen+1));
	WideCharToMultiByte( 932, 0, Item.strDevicePath.c_str(), -1, pszBuff, iLen, NULL, NULL );

	OutputLog("DevicePath:%s",pszBuff);
	delete[] pszBuff;
#endif
				}
				VariantClear(&varName);

				if( Item.strFriendlyName.compare(HDUS_SOURCE_NAME) == 0 && Item.strDevicePath.find(L"\\usb") != string::npos ){
					pDevList->push_back(Item);
				}
				pPropBag->Release();
			}
			pIMoniker.Release();
		}
		pIEnumMoniker.Release();
	}
	pICreateDevEnum.Release();
	return TRUE;
}

BOOL CDShowMain::IsEmptyTuner(wstring strDevPath)
{
	wstring strMutex = L"Global\\";
	int pos = (int)strDevPath.find(L"#");
	if( pos != string::npos ){
		pos = (int)strDevPath.find(L"#",pos+1);
	}
	int pos2 = (int)strDevPath.rfind(L"#");
	strMutex += strDevPath.substr(pos+1, pos2-(pos+1));
	HANDLE hMutex = CreateMutex( NULL,TRUE,strMutex.c_str() );
	if( hMutex == NULL ){
		return FALSE;
	}else if( ERROR_ALREADY_EXISTS == GetLastError()){
		::ReleaseMutex(hMutex);
		::CloseHandle(hMutex);
		return FALSE;
	}
	::ReleaseMutex(hMutex);
	::CloseHandle(hMutex);


	HRESULT hr = S_OK;
	CComPtr <IMoniker> pIMoniker;
	CComPtr <IEnumMoniker> pIEnumMoniker;
	CComPtr <ICreateDevEnum> pICreateDevEnum;

	REFCLSID clsid = KSCATEGORY_BDA_NETWORK_TUNER;

	hr = pICreateDevEnum.CoCreateInstance(CLSID_SystemDeviceEnum);
	if (FAILED (hr)){
		return FALSE;
	}
	BOOL bFind = FALSE;
	hr = pICreateDevEnum->CreateClassEnumerator(KSCATEGORY_BDA_NETWORK_TUNER, &pIEnumMoniker, 0);
	if (hr == S_OK){
		ULONG cFetched;
		while(pIEnumMoniker->Next(1, &pIMoniker, &cFetched) == S_OK){
			IPropertyBag *pPropBag;
			hr = pIMoniker->BindToStorage(0, 0, IID_IPropertyBag, 
				(void **)&pPropBag);
			if (SUCCEEDED(hr))
			{
				VARIANT varName;
				varName.vt = VT_BSTR;
				hr = pPropBag->Read(L"DevicePath", &varName, 0);
				if (SUCCEEDED(hr)){
					if( strDevPath.compare(varName.bstrVal) == 0 ){
						IBaseFilter *pFilter;
						HRESULT hr = pIMoniker->BindToObject(0, 0, IID_IBaseFilter, (void**)&pFilter);
						if (FAILED(hr)){
							return FALSE;
						}
						//Pin̊mF
						IEnumPins *pEnum = NULL;
						hr = pFilter->EnumPins(&pEnum);
						if (FAILED(hr)){
#ifdef _LOGOUT
		OutputLog("BDA_NETWORK_TUNER EnumPins Err 0x%x", hr);
#endif
							SAFE_RELEASE(pFilter);
							return FALSE;
						}
						ULONG ulFetched = 0;
						IPin *pPin = NULL;
						hr = pEnum->Next(1, &pPin, &ulFetched);
						if(SUCCEEDED(hr)){
							if( hr != S_FALSE){
								bFind = TRUE;
							}else{
#ifdef _LOGOUT
		OutputLog("BDA_NETWORK_TUNER EnumPins Next Err 0x%x", hr);
#endif
							}
						}
						SAFE_RELEASE(pEnum);
						SAFE_RELEASE(pPin);
						SAFE_RELEASE(pFilter);
					}
				}
				VariantClear(&varName);

				SAFE_RELEASE(pPropBag);
			}
			pIMoniker = NULL;
		}
	}

	return bFind;
}

BOOL CDShowMain::GetEmptyDevicePath(wstring& strDevPath)
{
	vector<DEVICE_INFO_DATA> DevList;
	if( EnumBDADevice(&DevList) == TRUE ){
		for( int i=0; i<(int)DevList.size(); i++ ){
			if( IsEmptyTuner(DevList[i].strDevicePath) == TRUE ){
				strDevPath = DevList[i].strDevicePath;
				return TRUE;
			}
		}
	}else{
#ifdef _LOGOUT
		OutputLog("EnumBDADevice Err");
#endif
		return FALSE;
	}
#ifdef _LOGOUT
		OutputLog("NoEmpty BDADevice");
#endif
	return FALSE;
}


// Adds a DirectShow filter graph to the Running Object Table,
// allowing GraphEdit to "spy" on a remote filter graph.
HRESULT CDShowMain::AddGraphToRot(
		IUnknown *pUnkGraph, 
		DWORD *pdwRegister
		) 
{
	CComPtr <IMoniker>              pMoniker;
	CComPtr <IRunningObjectTable>   pROT;
	WCHAR wsz[128];
	HRESULT hr;

	if (FAILED(GetRunningObjectTable(0, &pROT)))
		return E_FAIL;

	wsprintfW(wsz, L"FilterGraph %08x pid %08x\0", (DWORD_PTR) pUnkGraph, 
			GetCurrentProcessId());

	hr = CreateItemMoniker(L"!", wsz, &pMoniker);
	if (SUCCEEDED(hr)) 
		hr = pROT->Register(ROTFLAGS_REGISTRATIONKEEPSALIVE, pUnkGraph, 
							pMoniker, pdwRegister);
        
	return hr;
}

// Removes a filter graph from the Running Object Table
void CDShowMain::RemoveGraphFromRot(
		DWORD pdwRegister
		)
{
	CComPtr <IRunningObjectTable> pROT;

	if (SUCCEEDED(GetRunningObjectTable(0, &pROT))) 
		pROT->Revoke(pdwRegister);

}

BOOL CDShowMain::OpenTuner(wstring strDevPath, RECV_PROC pRecv, void* pParam)
{
	if( m_bCreate != FALSE ){
		CloseTuner();
	}
//	SAFE_DELETE(m_pcDmp);
//	SAFE_DELETE(m_pcTee);
#ifdef _LOGOUT
		OutputLog("OpenTuner" );
#endif
//	if(SetPriorityClass (GetCurrentProcess(), HIGH_PRIORITY_CLASS) == TRUE){
	HANDLE hCurThread = GetCurrentThread();
	if( SetThreadPriority(hCurThread, THREAD_PRIORITY_HIGHEST) == TRUE ){
#ifdef _LOGOUT
		OutputLog("Priority Change" );
#endif
	}

	HRESULT hr = S_OK;

	hr = m_pGraphBuilder.CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER);
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("FilterGraph CoCreateInstance Err 0x%x", hr);
#endif
		return FALSE;
	}

#ifdef _DEBUG
	AddGraphToRot(m_pGraphBuilder, &m_dwRegister);
#endif
	hr = m_pGraphBuilder->QueryInterface(IID_IMediaControl, reinterpret_cast<void**>(&m_pMediaControl));
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("IMediaControl QueryInterface Err 0x%x", hr);
#endif
		return FALSE;
	}

	hr = m_pTuningSpace.CoCreateInstance(CLSID_DVBTuningSpace);
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("DVBTuningSpace CoCreateInstance Err 0x%x", hr);
#endif
		return FALSE;
	}
	CComQIPtr <IDVBTuningSpace2> pDVBTuningSpace(m_pTuningSpace);
	if( !pDVBTuningSpace ){
		return FALSE;
	}
	hr = pDVBTuningSpace->put_SystemType(DVB_Terrestrial);
	if( FAILED(hr) ){
		return FALSE;
	}
	hr = pDVBTuningSpace->put__NetworkType(CLSID_DVBTNetworkProvider);
	if( FAILED(hr) ){
		return FALSE;
	}
	pDVBTuningSpace.Release();

	//tB^[쐬
	hr = CreateDeviceFilter(reinterpret_cast<IBaseFilter**>(&m_pNetworkProvider), KSCATEGORY_BDA_NETWORK_PROVIDER, DVBT_NP_NAME);
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("BDA_NETWORK_PROVIDER CreateDeviceFilter Err 0x%x", hr);
#endif
		return FALSE;
	}
	hr = m_pGraphBuilder->AddFilter(m_pNetworkProvider, DVBT_NP_NAME);
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("BDA_NETWORK_PROVIDER AddFilter Err 0x%x", hr);
#endif
		return FALSE;
	}

	CComPtr <ITuneRequest> pTuneRequest;
	hr = CreateTuneRequest(m_pTuningSpace, reinterpret_cast<ITuneRequest**>(&pTuneRequest), -1, -1 );
	if( FAILED(hr) ){
		return FALSE;
	}
	CComQIPtr <ITuner> pTuner(m_pNetworkProvider);
	if( !pTuner ){
		return FALSE;
	}
	hr = pTuner->put_TuneRequest(pTuneRequest);
	if( FAILED(hr) ){
		return FALSE;
	}
	pTuner.Release();

	hr = CreateAddTunerDevFilter(m_pGraphBuilder, (WCHAR*)strDevPath.c_str() );
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("TunerDevice->CaptureDevice CreateAddTunerDevFilter Err 0x%x", hr);
#endif
		return FALSE;
	}

	m_pcTee = static_cast<CTee *>(CTee::CreateInstance(NULL, &hr));
	m_pcTee->SetCallBackRecv(pRecv, pParam);
	hr = m_pcTee->QueryInterface(IID_IBaseFilter, reinterpret_cast<void**>(&m_pInfPinTee));
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("InfPinTee CreateDeviceFilter Err 0x%x", hr);
#endif
		return FALSE;
	}
	hr = m_pGraphBuilder->AddFilter(m_pInfPinTee, L"Infinite Pin Tee");
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("InfPinTee AddFilter Err 0x%x", hr);
#endif
		return FALSE;
	}
/*	m_pcDmp = static_cast<CDump *>(CDump::CreateInstance(NULL, &hr));
	m_pcDmp->SetCallBackRecv(pRecv, pParam);
	hr = m_pcDmp->QueryInterface(IID_IBaseFilter, reinterpret_cast<void**>(&m_pDump));
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("Dump CreateDeviceFilter Err 0x%x", hr);
#endif
		return FALSE;
	}
	hr = m_pGraphBuilder->AddFilter(m_pDump, L"Dump");
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("Dump AddFilter Err 0x%x", hr);
#endif
		return FALSE;
	}
*/	hr = CoCreateInstance(CLSID_MPEG2Demultiplexer, NULL, CLSCTX_INPROC_SERVER,IID_IBaseFilter, reinterpret_cast<void**>(&m_pMpeg2Demulti) );
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("MPEG2Demultiplexer CoCreateInstance Err 0x%x", hr);
#endif
		return FALSE;
	}
	hr = m_pGraphBuilder->AddFilter( m_pMpeg2Demulti, MPEG2_DEMULTI_NAME );
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("MPEG2Demultiplexer AddFilter Err 0x%x", hr);
#endif
		return FALSE;
	}
	hr = CreateDeviceFilter( reinterpret_cast<IBaseFilter**>(&m_pTIF), KSCATEGORY_BDA_TRANSPORT_INFORMATION, L"MPEG2 Transport Information Filter" );
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("BDA_TRANSPORT_INFORMATION CreateDeviceFilter Err 0x%x", hr);
#endif
		return FALSE;
	}
	hr = m_pGraphBuilder->AddFilter(m_pTIF, L"MPEG2 Transport Information Filter");
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("MPEG2TransportInformation AddFilter Err 0x%x", hr);
#endif
		return FALSE;
	}

#ifdef _LOGOUT
	OutputLog("NetworkProvider->TunerDevice ConnectFilters");
#endif
	hr = ConnectFilters(m_pGraphBuilder, m_pNetworkProvider, m_pTunerDevice);
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("NetworkProvider->TunerDevice ConnectFilters Err 0x%x", hr);
#endif
		return FALSE;
	}
//	ConnectFilters(m_pGraphBuilder, m_pTunerDevice, m_pCaptureDevice);
#ifdef _LOGOUT
	OutputLog("CaptureDevice->InfPinTee ConnectFilters");
#endif
	hr = ConnectFilters(m_pGraphBuilder, m_pCaptureDevice, m_pInfPinTee);
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("CaptureDevice->InfPinTee ConnectFilters Err 0x%x", hr);
#endif
		return FALSE;
	}
#ifdef _LOGOUT
	OutputLog("InfPinTee->Dump ConnectFilters ConnectFilters");
#endif
//	hr = ConnectFilters(m_pGraphBuilder, m_pInfPinTee, m_pDump);
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("InfPinTee->Dump ConnectFilters Err 0x%x", hr);
#endif
		return FALSE;
	}
#ifdef _LOGOUT
	OutputLog("InfPinTee->Mpeg2Demulti ConnectFilters");
#endif
	hr = ConnectFilters(m_pGraphBuilder, m_pInfPinTee, m_pMpeg2Demulti);
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("InfPinTee->Mpeg2Demulti ConnectFilters Err 0x%x", hr);
#endif
		return FALSE;
	}
#ifdef _LOGOUT
	OutputLog("Mpeg2Demulti->TIF ConnectFilters ConnectFilters");
#endif
	hr = ConnectFilters(m_pGraphBuilder, m_pMpeg2Demulti, m_pTIF);
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("Mpeg2Demulti->TIF ConnectFilters Err 0x%x", hr);
#endif
		return FALSE;
	}


	m_pMediaControl->Run();
#ifdef _LOGOUT
	OutputLog("Run");
#endif

	m_bCreate = TRUE;
	m_iLastChFreq = -1;
	return TRUE;

}

BOOL CDShowMain::CloseTuner()
{
	_set_se_translator(err_translator);
	if( m_bCreate == FALSE ){
		return FALSE;
	}

	m_bCreate = FALSE;
	try{
		m_pMediaControl->Pause();
		m_pMediaControl->Stop();
	}catch(...){
	}
	m_pMediaControl = NULL;

#ifdef _DEBUG
	if( m_bCreate != FALSE ){
		RemoveGraphFromRot(m_dwRegister);
	}
#endif
	m_pTuningSpace = NULL;

	m_pGraphBuilder->RemoveFilter(m_pTIF);
	m_pGraphBuilder->RemoveFilter(m_pMpeg2Demulti);
//	m_pGraphBuilder->RemoveFilter(m_pDump);
	m_pGraphBuilder->RemoveFilter(m_pInfPinTee);
	m_pGraphBuilder->RemoveFilter(m_pNetworkProvider);
	m_pGraphBuilder->RemoveFilter(m_pTunerDevice);
	m_pGraphBuilder->RemoveFilter(m_pCaptureDevice);

	m_pNetworkProvider = NULL;
	m_pTunerDevice = NULL;
	m_pCaptureDevice = NULL;
	m_pInfPinTee = NULL;
//	m_pDump = NULL;
	m_pMpeg2Demulti = NULL;
	m_pTIF = NULL;
	m_pGraphBuilder = NULL;

//	m_pcDmp = NULL;
	m_pcTee = NULL;

//	if(SetPriorityClass (GetCurrentProcess(), NORMAL_PRIORITY_CLASS) == TRUE){
	HANDLE hCurThread = GetCurrentThread();
	if( SetThreadPriority(hCurThread, THREAD_PRIORITY_NORMAL) == TRUE ){
#ifdef _LOGOUT
		OutputLog("Priority Change" );
#endif
	}

#ifdef _LOGOUT
		OutputLog("CloseTuner" );
#endif
	return TRUE;
}

BOOL CDShowMain::SetChFreq(int iFreq)
{
	if( m_bCreate == FALSE ){
		return FALSE;
	}
	if( m_iLastChFreq == iFreq ){
#ifdef _LOGOUT
		OutputLog("Last Freq == Set Freq" );
#endif
		return TRUE;
	}
#ifdef _LOGOUT
		OutputLog("SetChFreq" );
#endif
	m_pMediaControl->Stop();
	for( int i=0; i<10; i++ ){
		CComQIPtr <ITuner> pTuner(m_pNetworkProvider);
		if( !pTuner ){
			continue;
		}

		CComPtr <ITuneRequest> pTuneRequest;
		HRESULT hr = CreateTuneRequest(m_pTuningSpace, reinterpret_cast<ITuneRequest**>(&pTuneRequest), iFreq, 6);
		if (SUCCEEDED(hr)){
#ifdef _LOGOUT
		OutputLog("put_TuneRequest" );
#endif
			hr = pTuner->put_TuneRequest(pTuneRequest);
			pTuneRequest.Release();
			pTuner.Release();
			break;
		}else{
			pTuner.Release();
			Sleep(100);
		}
	}
	m_pMediaControl->Run();
#ifdef _LOGOUT
		OutputLog("SetChFreq End" );
#endif
	m_iLastChFreq = iFreq;

	return TRUE;
}

// ConnectFilters is called from BuildGraph
// to enumerate and connect pins
HRESULT
CDShowMain::ConnectFilters(
	IGraphBuilder* pGraphBuilder,
	IBaseFilter* pFilterUpstream, 
	IBaseFilter* pFilterDownstream
    )
{
    HRESULT         hr = E_FAIL;

    CComPtr <IPin>  pIPinUpstream;


    PIN_INFO        PinInfoUpstream;
    PIN_INFO        PinInfoDownstream;

    // validate passed in filters
    ASSERT (pFilterUpstream);
    ASSERT (pFilterDownstream);

    // grab upstream filter's enumerator
    CComPtr <IEnumPins> pIEnumPinsUpstream;
    hr = pFilterUpstream->EnumPins(&pIEnumPinsUpstream);

    if(FAILED(hr))
    {
        return hr;
    }

    // iterate through upstream filter's pins
    while (pIEnumPinsUpstream->Next (1, &pIPinUpstream, 0) == S_OK)
    {
        hr = pIPinUpstream->QueryPinInfo (&PinInfoUpstream);
        if(FAILED(hr))
        {
            return hr;
        }

        CComPtr <IPin>  pPinDown;
        pIPinUpstream->ConnectedTo (&pPinDown);

        // bail if pins are connected
        // otherwise check direction and connect
        if ((PINDIR_OUTPUT == PinInfoUpstream.dir) && (pPinDown == NULL))
        {
            // grab downstream filter's enumerator
            CComPtr <IEnumPins> pIEnumPinsDownstream;
            hr = pFilterDownstream->EnumPins (&pIEnumPinsDownstream);
            if(FAILED(hr))
            {
                return hr;
            }

            // iterate through downstream filter's pins
            CComPtr <IPin>  pIPinDownstream;
            while (pIEnumPinsDownstream->Next (1, &pIPinDownstream, 0) == S_OK)
            {
                // make sure it is an input pin
                hr = pIPinDownstream->QueryPinInfo(&PinInfoDownstream);
                if(SUCCEEDED(hr))
                {
                    CComPtr <IPin>  pPinUp;

                    // Determine if the pin is already connected.  Note that 
                    // VFW_E_NOT_CONNECTED is expected if the pin isn't yet connected.
                    hr = pIPinDownstream->ConnectedTo (&pPinUp);
                    if(FAILED(hr) && hr != VFW_E_NOT_CONNECTED)
                    {
                        continue;
                    }

                    if ((PINDIR_INPUT == PinInfoDownstream.dir) && (pPinUp == NULL))
                    {
						if (SUCCEEDED (pGraphBuilder->Connect(
                                        pIPinUpstream,
                                        pIPinDownstream))
                                        )
                        {
                            PinInfoDownstream.pFilter->Release();
                            PinInfoUpstream.pFilter->Release();
                            return S_OK;
                        }
                    }
                }

                PinInfoDownstream.pFilter->Release();
                pIPinDownstream = NULL;
            } // while next downstream filter pin

            //We are now back into the upstream pin loop
        } // if output pin

        pIPinUpstream = NULL;
        PinInfoUpstream.pFilter->Release();
    } // while next upstream filter pin

    return E_FAIL;
}

HRESULT CDShowMain::CreateTuneRequest(ITuningSpace* pTuningSpace, ITuneRequest** ppTuneRequest, LONG lFreq, LONG lBandWidth)
{
	HRESULT hr = S_OK;
	if( pTuningSpace == NULL ){
		return E_FAIL;
	}
	CComQIPtr <IDVBTuningSpace2> pDVBTuningSpace(pTuningSpace);
	if( !pDVBTuningSpace ){
		return E_FAIL;
	}

	hr = pDVBTuningSpace->CreateTuneRequest(ppTuneRequest);
	if (FAILED(hr)){
		return hr;
	}
	CComQIPtr <IDVBTuneRequest> piDVBTuneRequest(*ppTuneRequest);
	if (!piDVBTuneRequest){
		pDVBTuningSpace.Release();
		return hr;
	}

	CComPtr <IDVBTLocator> pDVBTLocator;
	hr = pDVBTLocator.CoCreateInstance(CLSID_DVBTLocator);
	if (FAILED(hr)){
		pDVBTuningSpace.Release();
		piDVBTuneRequest.Release();
		SAFE_RELEASE(*ppTuneRequest);
		return hr;
	}
	hr = pDVBTLocator->put_CarrierFrequency(lFreq);
	if (FAILED(hr)){
		pDVBTuningSpace.Release();
		piDVBTuneRequest.Release();
		SAFE_RELEASE(*ppTuneRequest);
		return hr;
	}
	hr = pDVBTLocator->put_Bandwidth(lBandWidth);
	if (FAILED(hr)){
		pDVBTuningSpace.Release();
		piDVBTuneRequest.Release();
		SAFE_RELEASE(*ppTuneRequest);
		return hr;
	}
	hr = (*ppTuneRequest)->put_Locator(pDVBTLocator);

	pDVBTuningSpace.Release();
	piDVBTuneRequest.Release();
	return hr;
}

HRESULT CDShowMain::CreateDeviceFilter(
	IBaseFilter **ppFilter,
	REFCLSID	rciCategory,
	WCHAR		*pwzName
)
{
	HRESULT	hr		  = S_OK;
	ULONG	ulFetched = 0;
	BOOL	bFound	  = FALSE;

	ICreateDevEnum *pDevEnum = NULL;
	IEnumMoniker   *pEnum	 = NULL;
	IMoniker	   *pMoniker = NULL;
	IPropertyBag   *pBag	 = NULL;

	VARIANT varName;
	varName.vt = VT_BSTR;

	if (!ppFilter || !pwzName) return E_POINTER;

	*ppFilter = NULL;

	// VXefoCX񋓎q쐬
	hr = CoCreateInstance( CLSID_SystemDeviceEnum,
						   NULL,
						   CLSCTX_INPROC,
						   IID_ICreateDevEnum,
						   (void**) &pDevEnum );

    if (FAILED(hr)) goto Err_CPF;

	// JeS񋓂NX񋓎q쐬
	hr = pDevEnum->CreateClassEnumerator(rciCategory, &pEnum, 0);

    if (FAILED(hr)) goto Err_CPF;

	if (!pEnum)
	{
		hr = E_FAIL;
		goto Err_CPF;
	}

	// JeSɊ܂܂foCX
	while(pEnum->Next(1, &pMoniker, &ulFetched) == S_OK)
	{
		hr = pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void**) &pBag);
		if (SUCCEEDED(hr)){
			// th擾
			hr = pBag->Read(L"FriendlyName", &varName, NULL);
			if (SUCCEEDED(hr)){

				// mF
				if (wcsstr(varName.bstrVal, pwzName) != NULL ){
					// tB^̂ŃtB^擾
					hr = pMoniker->BindToObject(0, 0, IID_IBaseFilter, (void**) ppFilter);
					if (SUCCEEDED(hr)){
						bFound = TRUE;
					}
				}
				SysFreeString(varName.bstrVal);
			}

			SAFE_RELEASE(pBag);
		}
		SAFE_RELEASE(pMoniker);

		if (bFound != FALSE){
			break;
		}else{
			SAFE_RELEASE(*ppFilter);
		}
	}

	SAFE_RELEASE(pEnum);
	SAFE_RELEASE(pDevEnum);

	return S_OK;

Err_CPF:

	SAFE_RELEASE(pBag);
	SAFE_RELEASE(pMoniker);
	SAFE_RELEASE(pEnum);
	SAFE_RELEASE(pDevEnum);

	return hr;
}

HRESULT CDShowMain::CreateDeviceFilter(
	IBaseFilter **ppFilter,
	REFCLSID	rciCategory,
	WCHAR		*pwzName,
	WCHAR		*pwzDevPath
)
{
	HRESULT	hr		  = S_OK;
	ULONG	ulFetched = 0;
	BOOL	bFound	  = FALSE;

	ICreateDevEnum *pDevEnum = NULL;
	IEnumMoniker   *pEnum	 = NULL;
	IMoniker	   *pMoniker = NULL;
	IPropertyBag   *pBag	 = NULL;

	VARIANT varName;
	varName.vt = VT_BSTR;
	VARIANT varPath;
	varPath.vt = VT_BSTR;

	if (!ppFilter || !pwzName) return E_POINTER;

	*ppFilter = NULL;

	// VXefoCX񋓎q쐬
	hr = CoCreateInstance( CLSID_SystemDeviceEnum,
						   NULL,
						   CLSCTX_INPROC,
						   IID_ICreateDevEnum,
						   (void**) &pDevEnum );

    if (FAILED(hr)) goto Err_CPF;

	// JeS񋓂NX񋓎q쐬
	hr = pDevEnum->CreateClassEnumerator(rciCategory, &pEnum, 0);

    if (FAILED(hr)) goto Err_CPF;

	if (!pEnum)
	{
		hr = E_FAIL;
		goto Err_CPF;
	}

	// JeSɊ܂܂foCX
	while(pEnum->Next(1, &pMoniker, &ulFetched) == S_OK)
	{
		hr = pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void**) &pBag);
		if (SUCCEEDED(hr)){
			// th擾
			hr = pBag->Read(L"FriendlyName", &varName, NULL);
			if (SUCCEEDED(hr)){
				//foCXpX
				hr = pBag->Read(L"DevicePath", &varPath, NULL);
				if (SUCCEEDED(hr)){

					// mF
					if (wcsstr(varName.bstrVal, pwzName) != NULL && wcsstr(varPath.bstrVal, pwzDevPath) != NULL ){
						// tB^̂ŃtB^擾
						hr = pMoniker->BindToObject(0, 0, IID_IBaseFilter, (void**) ppFilter);
						if (SUCCEEDED(hr)){
							bFound = TRUE;

						}
					}
					SysFreeString(varName.bstrVal);
				}
				SysFreeString(varPath.bstrVal);
			}
			SAFE_RELEASE(pBag);
		}

		SAFE_RELEASE(pMoniker);

		if (bFound != FALSE){
			break;
		}else{
			SAFE_RELEASE(*ppFilter);
		}
	}

	SAFE_RELEASE(pEnum);
	SAFE_RELEASE(pDevEnum);

	return S_OK;

Err_CPF:

	SAFE_RELEASE(pBag);
	SAFE_RELEASE(pMoniker);
	SAFE_RELEASE(pEnum);
	SAFE_RELEASE(pDevEnum);

	return hr;
}

HRESULT CDShowMain::CreateAddTunerDevFilter(
	IGraphBuilder* pGraphBuilder,
	WCHAR		*pwzDevPath
)
{
	//`[i[̓foCXpXŌĕʂɒǉ
	HRESULT hr = CreateDeviceFilter( reinterpret_cast<IBaseFilter**>(&m_pTunerDevice), KSCATEGORY_BDA_NETWORK_TUNER, HDUS_SOURCE_NAME, pwzDevPath );
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("BDA_NETWORK_TUNER CreateDeviceFilter Err 0x%x", hr);
#endif
		return hr;
	}
	hr = m_pGraphBuilder->AddFilter(m_pTunerDevice, HDUS_SOURCE_NAME);
	if( FAILED(hr) ){
#ifdef _LOGOUT
		OutputLog("BDA_NETWORK_TUNER AddFilter Err 0x%x", hr);
#endif
		return hr;
	}

	//V[o͂Ȃ̂T
	return CreateAddReciverFilter(pGraphBuilder, m_pTunerDevice, reinterpret_cast<IBaseFilter**>(&m_pCaptureDevice), KSCATEGORY_BDA_RECEIVER_COMPONENT, HDUS_COMPONENT_NAME );
}

HRESULT CDShowMain::CreateAddReciverFilter(
	IGraphBuilder* pGraphBuilder,
	IBaseFilter *pTunerFilter,
	IBaseFilter **ppFilter,
	REFCLSID	rciCategory,
	WCHAR		*pwzName
)
{
	HRESULT	hr		  = S_OK;
	ULONG	ulFetched = 0;
	BOOL	bFound	  = FALSE;

	ICreateDevEnum *pDevEnum = NULL;
	IEnumMoniker   *pEnum	 = NULL;
	IMoniker	   *pMoniker = NULL;
	IPropertyBag   *pBag	 = NULL;

	VARIANT varName;
	varName.vt = VT_BSTR;

	if (!ppFilter || !pwzName) return E_POINTER;

	*ppFilter = NULL;

	// VXefoCX񋓎q쐬
	hr = CoCreateInstance( CLSID_SystemDeviceEnum,
						   NULL,
						   CLSCTX_INPROC,
						   IID_ICreateDevEnum,
						   (void**) &pDevEnum );

    if (FAILED(hr)) goto Err_CPF;

	// JeS񋓂NX񋓎q쐬
	hr = pDevEnum->CreateClassEnumerator(rciCategory, &pEnum, 0);

    if (FAILED(hr)) goto Err_CPF;

	if (!pEnum)
	{
		hr = E_FAIL;
		goto Err_CPF;
	}

	// JeSɊ܂܂foCX
	while(pEnum->Next(1, &pMoniker, &ulFetched) == S_OK)
	{
		hr = pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void**) &pBag);
		if (SUCCEEDED(hr)){
			// th擾
			hr = pBag->Read(L"FriendlyName", &varName, NULL);
			if (SUCCEEDED(hr)){
				// mF
				if (wcsstr(varName.bstrVal, pwzName) != NULL ){
					// tB^̂ŃtB^擾
					hr = pMoniker->BindToObject(0, 0, IID_IBaseFilter, (void**) ppFilter);
					if (SUCCEEDED(hr)){
						hr = m_pGraphBuilder->AddFilter(*ppFilter, pwzName);
						if( SUCCEEDED(hr) ){
							hr = ConnectFilters(m_pGraphBuilder, pTunerFilter, *ppFilter);
							if( SUCCEEDED(hr) ){
								bFound = TRUE;
							}else{
								m_pGraphBuilder->RemoveFilter(*ppFilter);
							}
						}
					}
				}
				SysFreeString(varName.bstrVal);
			}
			SAFE_RELEASE(pBag);
		}

		SAFE_RELEASE(pMoniker);

		if (bFound != FALSE){
			break;
		}else{
			SAFE_RELEASE(*ppFilter);
		}
	}

	SAFE_RELEASE(pEnum);
	SAFE_RELEASE(pDevEnum);

	if (bFound != FALSE){
		return S_OK;
	}else{
		return E_FAIL;
	}

Err_CPF:

	SAFE_RELEASE(pBag);
	SAFE_RELEASE(pMoniker);
	SAFE_RELEASE(pEnum);
	SAFE_RELEASE(pDevEnum);

	return hr;
}

void CDShowMain::OutputLog(const char* log, ...)
{
	char buffer[2048] = "";
	va_list ap;
	va_start(ap, log);
	vsprintf(buffer, log, ap);
	va_end(ap);

	SYSTEMTIME stNow;
	GetLocalTime(&stNow);
	
	char strTime[256]="";
	sprintf(strTime, "%04d/%02d/%02d %02d:%02d:%02d.%03d:",
		stNow.wYear,
		stNow.wMonth,
		stNow.wDay,
		stNow.wHour,
		stNow.wMinute,
		stNow.wSecond,
		stNow.wMilliseconds
		);

	string strOut = "";
	strOut = strTime;
	strOut+=buffer;
	strOut+="\r\n";

	HANDLE hFile = CreateFileW( m_strLogPath.c_str(), GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
	if( hFile == INVALID_HANDLE_VALUE ){
		return ;
	}
	SetFilePointer(hFile, 0, NULL, FILE_END);
	DWORD dwWrite=0;
	WriteFile(hFile, strOut.c_str(), (DWORD)strOut.length(), &dwWrite, NULL );
	CloseHandle(hFile);
}


int CDShowMain::get_level(DWORD param)
{
  HRESULT hr;
  IKsPropertySet *pIKsPropertySet;
  BYTE buf[4];
  DWORD m;

  hr = m_pTunerDevice->QueryInterface(
    IID_IKsPropertySet,(LPVOID*)&pIKsPropertySet);
  if (hr != S_OK) {
//    fprintf(stderr,"get_level: QueryInterface\n");
    return -2;
  }

  hr = pIKsPropertySet->Set( 
    CLSID_PropSet,3,NULL,0,(void*)&param,4);
  if (hr != S_OK) {
//    fprintf(stderr,"get_level: IKsPropertySet.Set hr=%08X\n",hr);
    goto get_level_last;
  }

  hr = pIKsPropertySet->Get(
    CLSID_PropSet,3,&param,4,&buf,4,&m);
  if (hr == S_OK && m == 4) {
    pIKsPropertySet->Release();
    return buf[0];
  }

get_level_last:
  param = 1;
  hr = pIKsPropertySet->Set(
    CLSID_PropSet,4,NULL,0,&param,4);
  pIKsPropertySet->Release();
  if (hr != S_OK) {
//    fprintf(stderr,"get_level: IKsPropertySet.Set (2) hr=%08X\n",hr);
  }
  return -2;
}

double CDShowMain::get_level1(void)
{
  int m,n;
  double r;

  if (m_bCreate == FALSE)
    return -2;
  m = get_level(0x8b20);
  if (m < 0)
    return m;
  n = m << 8;
  m = get_level(0x8c20);
  if (m < 0)
    return m;
  n += m;
  m = get_level(0x8d20);
  if (m < 0)
    return m;

  n = (n << 8)+m;
  if (n == 0)
    return -1;
  r = 20*log10(5505024.0/n);
  if (r > 99)
    return 99;
  return r;
}

double CDShowMain::get_level2(void)
{
  int m,n;
  double r;

  if (m_bCreate == FALSE)
    return -2;
  m = get_level(0xa020);
  if (m < 0)
    return m;
  n = m << 8;
  m = get_level(0xa120);
  if (m < 0)
    return m;
  n += m;
  m = get_level(0xa220);
  if (m < 0)
    return m;
  r = (n << 8)+m;

  m = get_level(0xa620);
  if (m < 0)
    return m;
  n = m << 8;
  m = get_level(0xa720);
  if (m < 0)
    return m;
  return r/((n+m)*1632.0);
}

BOOL CDShowMain::GetLevel(double* level1, double* level2)
{
	if(m_bCreate == TRUE){
		if( level1 != NULL ){
			*level1 = get_level1();
		}
		if( level2 != NULL ){
			*level2 = get_level2();
		}
	}else{
		if( level1 != NULL ){
			*level1 = 0;
		}
		if( level2 != NULL ){
			*level2 = 0;
		}
	}
	return TRUE;
}
