#pragma once

#include <streams.h>
#include <mmreg.h>
#include <msacm.h>
#include <fcntl.h>
#include <io.h>
#include <stdio.h>
#include <ks.h>
#include <ksmedia.h>
#include <bdatypes.h>
#include <bdamedia.h>
#include <bdaiface.h>
#include <uuids.h>
#include <tuner.h>
#include <commctrl.h>
#include <atlbase.h>
#include <refclock.h>

#include "Dump.h"
#include "inftee.h"

#ifdef _DEBUG
#pragma comment(lib, "strmbasd.lib")
#else
#pragma comment(lib, "strmbase.lib")
#endif

#include <vector>
#include <algorithm>
#include <string>
using namespace std;

#define SAFE_RELEASE(p)      { if(p) { (p)->Release(); (p)=NULL; } }
#define SAFE_DELETE(p)       { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_DELETE_ARRAY(p) { if(p) { delete[] (p);   (p)=NULL; } }

class CDShowMain
{
public:
	typedef struct _DEVICE_INFO_DATA{
		wstring strFriendlyName;
		wstring strDescription;
		wstring strDevicePath;
	}DEVICE_INFO_DATA;

public:
	CDShowMain(void);
	~CDShowMain(void);

	void SetModulePath(wstring strModulePath);
	BOOL GetEmptyDevicePath(wstring& strDevPath);

	BOOL OpenTuner(wstring strDevPath, RECV_PROC pRecv, void* pParam);
	BOOL CloseTuner();

	BOOL SetChFreq(int iFreq);

	BOOL GetLevel(double* level1, double* level2);

	void OutputLog(const char* log, ...);

protected:
	DWORD m_dwRegister;
	BOOL m_bCreate;

	CComPtr <IGraphBuilder> m_pGraphBuilder;
	CComPtr <IMediaControl> m_pMediaControl;

	CComPtr <IBaseFilter> m_pNetworkProvider;
	CComPtr <IBaseFilter> m_pTunerDevice;
	CComPtr <IBaseFilter> m_pCaptureDevice;
	CComPtr <IBaseFilter> m_pInfPinTee;
//	CComPtr <IBaseFilter> m_pDump;
	CComPtr <IBaseFilter> m_pMpeg2Demulti;
	CComPtr <IBaseFilter> m_pTIF;

	CComPtr <ITuningSpace> m_pTuningSpace;

//	CDump* m_pcDmp;
	CTee* m_pcTee;

	wstring m_strLogPath;

	int m_iLastChFreq;

protected:
	BOOL EnumBDADevice(vector<DEVICE_INFO_DATA>* pDevList);
	BOOL IsEmptyTuner(wstring strDevPath);

	HRESULT AddGraphToRot(
		IUnknown *pUnkGraph, 
		DWORD *pdwRegister
		);

	void RemoveGraphFromRot(
		DWORD pdwRegister
		);

    HRESULT ConnectFilters(
		IGraphBuilder* pGraphBuilder,
        IBaseFilter* pFilterUpstream, 
        IBaseFilter* pFilterDownstream
        );

	HRESULT CreateTuneRequest(
		ITuningSpace* pTuningSpace,
		ITuneRequest** ppTuneRequest,
		LONG lFreq, LONG lBandWidth
		);

	HRESULT CreateDeviceFilter(
		IBaseFilter **ppFilter,
		REFCLSID	rciCategory,
		WCHAR		*pwzName
	);
	HRESULT CreateDeviceFilter(
		IBaseFilter **ppFilter,
		REFCLSID	rciCategory,
		WCHAR		*pwzName,
		WCHAR		*pwzDevPath
	);

	HRESULT CreateAddTunerDevFilter(
		IGraphBuilder* pGraphBuilder,
		WCHAR		*pwzDevPath
	);
	HRESULT CreateAddReciverFilter(
		IGraphBuilder* pGraphBuilder,
		IBaseFilter *pTunerFilter,
		IBaseFilter **ppFilter,
		REFCLSID	rciCategory,
		WCHAR		*pwzName
	);

	int get_level(DWORD param);
	double get_level1(void);
	double get_level2(void);
};
